/************************************************************

     display_text.c  --  Simple program to demonstrate how to
	use Xt's API for selections.

Copyright 1993, Hewlett-Packard Company.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and to
use, copy, and distribute its documentation for any purpose and without
fee is hereby granted, provided that the above copyright notice appear
in all copies and that both that copyright notice and this permission
notice appear in supporting documentation, and that the name of
Hewlett-Packard not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior permission.

HEWLETT-PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL HEWLETT-PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

******************************************************************/
#include <stdio.h>

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/Xmu/StdSel.h>

#include <Xm/Xm.h>
#include <Xm/Protocols.h>
#include <Xm/MainW.h>
#include <Xm/DrawingA.h>
#include <Xm/MessageB.h>

#define MAX_ARGS	25
#define MAX_TEXT_LEN	500

#define MIN(x, y)	((x > y) ? y : x)
#define MAX(x, y)	((x > y) ? x : y)

void		quitCB(), exposeCB();
Boolean		selection_requestCB(); 
void		lost_ownershipCB();
void		draw_char(), draw_range(), create_gcs();
void		start_select(), drag_select(), end_select();
void		paste_primary();
void		targetCB(), valueCB();
Atom		pick_best_target();
void		error_dialog();


static XtActionsRec actions[] = {
	{"StartSelect", start_select},
	{"DragSelect", drag_select},
	{"EndSelect", end_select},
	{"PastePrimary", paste_primary},
};

static char newTranslations[] =
	"<Btn1Down>:	StartSelect()\n\
	 <Btn1Motion>:	DragSelect()\n\
	 <Btn1Up>:	EndSelect()\n\
	 <Btn2Down>:	PastePrimary()";

char		*progname;
char		the_text[MAX_TEXT_LEN+1];
Dimension	win_width, win_height;
int		char_width, char_height;
Boolean		own_primary = False;
int		first_select_char = -1, 
		last_select_char = -1; 

Display		*dpy;
int		screen;
XtAppContext	app_context;
Widget		toplevel;
Atom      	xa_WM_DELETE_WINDOW;
Atom		xa_PRIMARY;
Atom		xa_TEXT;
Atom		xa_STRING;
Atom		xa_TARGETS;
Atom		xa_ATOM;
Atom		xa_LENGTH;
Atom		xa_INTEGER;
GC              normal_gc = NULL, inverse_gc;
char     	*fontname = "9x15";    /* Should be a resource ??? */
XFontStruct     *fontinfo;


void main (argc,argv) 
unsigned int	argc;
char 		**argv;
{
    Widget      main_win;
    Widget 	canvas;
    Arg         arg[MAX_ARGS];
    int         n;
    XtTranslations	my_translations;
    
    progname = argv[0];

    /*
       Initialize the toolkit, open the display, and create a 
       toplevel window.
    */
    XtToolkitInitialize();
    app_context = XtCreateApplicationContext();
    dpy = XtOpenDisplay (app_context, NULL, argv[0], "DisplayText", NULL, 0,
				&argc, argv);
    if (!dpy) {
	XtWarning ("display_text: Can't open display, exiting...");
	exit (0);
    }
    screen = DefaultScreen(dpy);

    if (argc > 1) {
	strncpy(the_text, argv[1], MAX_TEXT_LEN);
    } else {
	strcpy(the_text, "This is a default string because you didn't specify one on the command line");
    }

    n = 0;
    XtSetArg(arg[n], XmNdeleteResponse, XmDO_NOTHING); n++;
    XtSetArg(arg[n], XmNgeometry, "=300x100"); n++;
    toplevel = XtAppCreateShell (argv[0], "SelectOMatic",
				applicationShellWidgetClass, dpy, arg, n);

    /*
       Intern some atoms we'll be needing
    */
    xa_WM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    xa_PRIMARY = XInternAtom(dpy, "PRIMARY", False);
    xa_STRING = XInternAtom(dpy, "STRING", False);
    xa_TEXT = XInternAtom(dpy, "TEXT", False);
    xa_TARGETS = XInternAtom(dpy, "TARGETS", False);
    xa_LENGTH = XInternAtom(dpy, "LENGTH", False);
    xa_ATOM = XInternAtom(dpy, "ATOM", False);
    xa_INTEGER = XInternAtom(dpy, "INTEGER", False);

    /*
       Set it up so we find out when the window's closed
    */
    XmAddWMProtocolCallback(toplevel, xa_WM_DELETE_WINDOW, quitCB, NULL);

    n = 0;
    main_win = XmCreateMainWindow(toplevel, "main", arg, n);
    XtManageChild(main_win);

    /*
	Create the Drawing Area for displaying the string,
	and add our selection-related translations.
    */
    n = 0;
    canvas = XmCreateDrawingArea(main_win, "canvas", arg, n);
    XtManageChild(canvas);
    XtAddCallback(canvas, XmNexposeCallback, exposeCB, NULL);

    XtAppAddActions(app_context, actions, XtNumber(actions));
    my_translations = XtParseTranslationTable(newTranslations);
    XtSetArg(arg[0], XmNtranslations, my_translations);
    XtSetValues(canvas, arg, 1);

    XmMainWindowSetAreas(main_win, NULL, NULL, NULL, NULL, canvas);

    XtRealizeWidget (toplevel);
    
#ifdef DEBUG
    XtAppSetSelectionTimeout(app_context, 200000);
#endif

    XtAppMainLoop (app_context);
}


/*
	quitCB			- callback for quit button
*/
void quitCB (w, client_data, call_data) 
Widget		w;		/*  widget id		*/
caddr_t		client_data;	/*  data from application   */
caddr_t		call_data;	/*  data from widget class  */
{

    /*	
	Terminate the application.
     */
    exit (0);
}



/************************************************************

	     Stuff for Displaying the String

************************************************************/


void exposeCB (canvas, client_data, call_data) 
Widget canvas;
XtPointer client_data;
XtPointer call_data;
{
    Arg      arg[4];
    Window   w = XtWindow(canvas);

    /*
	First time through, set some things up
    */
    if (!normal_gc) {
        Pixel	    fg, bg;

	setup_gcs(XtWindow(canvas));   /* Assuming it works ??? */
    	XtSetArg(arg[0], XmNbackground, &bg);
    	XtSetArg(arg[1], XmNforeground, &fg);
    	XtGetValues(canvas, arg, 2);
    	XSetBackground(dpy, normal_gc, bg);
    	XSetForeground(dpy, normal_gc, fg);
    	XSetBackground(dpy, inverse_gc, fg);
    	XSetForeground(dpy, inverse_gc, bg);
        XtAddCallback(canvas, XmNresizeCallback, exposeCB, NULL);
    }

    /*
	Update our globals with the new width and height.
    */
    XtSetArg(arg[0], XmNwidth, &win_width); 
    XtSetArg(arg[1], XmNheight, &win_height);
    XtGetValues(canvas, arg, 2);

    /*
	Refresh the string. This is *not* the most efficient way
	to do this, but it was easy to code.
    */
    XClearWindow(dpy, w);


    if (first_select_char != last_select_char) {
    	draw_range(w, 0, first_select_char, normal_gc);
    	draw_range(w, first_select_char, last_select_char, inverse_gc);
        draw_range(w, last_select_char, strlen(the_text), normal_gc);
    } else {
	draw_range(w, 0, strlen(the_text), normal_gc);
    }
}


/*
    Draw the nth character of the string in Window w with GC gc.
*/

void draw_char (w, n, gc)
Window    w;
int       n;
GC	  gc;
{
    int   numcols;
    int   x, y;

    if ((n >= 0) && (n < strlen(the_text))) {
    	numcols = win_width / char_width;
	/*
    	x = ((n % numcols) * char_width) + fontinfo->max_bounds.lbearing;
    	y = ((n / numcols) * char_height) + fontinfo->max_bounds.ascent;
	*/
    	x = ((n % numcols) * char_width);
    	y = ((n / numcols) * char_height) + char_height;
    	XDrawImageString(dpy, w, gc, x, y, the_text + n, 1);
    }
}


/*
    Draw the characters char1 through char2 in Window w with GC gc.
    This function will handle the args in any order, so the drawing
    always includes the character with the lower index, but excludes
    the character with the higher index.
*/

void draw_range (w, char1, char2, gc)
Window   w;
int      char1, char2;
GC       gc;
{
    int  tmp, i;

    if (char1 > char2) {
	tmp = char1;
	char1 = char2;
	char2 = tmp;
    }

    for (i = char1; i < char2; i++)
	draw_char(w, i, gc);
}



/*
    Open the font we'll be using and create a couple gc's.
*/
	
setup_gcs (w)
Window   w;
{
    fontinfo = XLoadQueryFont(dpy, fontname);

    if (!fontinfo) {
       fprintf(stderr, "Unable to open font %s\n", fontname);
       exit (1); /* ??? */
    }
    char_height = fontinfo->max_bounds.ascent + fontinfo->max_bounds.descent;
    char_width = fontinfo->max_bounds.width;

    normal_gc = XCreateGC(dpy, w, 0l, NULL);
    XSetFont(dpy, normal_gc, fontinfo->fid);
    inverse_gc = XCreateGC(dpy, w, 0l, NULL);
    XSetFont(dpy, inverse_gc, fontinfo->fid);
}

    

/**************************************************************

   User Interface Stuff for Selecting Text (highlighting, etc.)

**************************************************************/


/*
    Translate a pixel coordinate into the index of the character
    in the string that appears at that pixel coordinate.
*/

int pixel_coords_to_char (x, y)
int   x, y;
{
    int  char_x, char_y;
    int  num_cols, num_rows;
    int  length;
    int  n;

    /*
	Do some range filtering
    */
    length = strlen(the_text);
    num_cols = win_width / char_width;
    num_rows = (length / num_cols) + 1;
    char_x = x / char_width;
    char_x = MAX(char_x, 0);
    char_x = MIN(char_x, num_cols);
    char_y = y / char_height;
    char_y = MAX(char_y, 0);
    char_y = MIN(char_y, num_rows);

    n = (char_y * num_cols) + char_x;

    if (n > length)
	n = length;

    return (n);
}


static int   done_first = False;
static int   last_drawn = -1;


/*
    The user has pressed Btn1. Remember the character the button
    was depressed over.
*/

void start_select (w, event)
Widget   w;
XButtonEvent  *event;
{

    /* 
       Erase the old selection highlight 
    */
    draw_range(XtWindow(w), first_select_char, last_select_char, normal_gc);

    done_first = False;
    first_select_char = pixel_coords_to_char (event->x, event->y);
    last_drawn = -1;
}


/*
    Highlight the text the user is selecting appropriately.
    Don't worry, even I'm not sure why it works.
*/

void drag_select (w, event)
Widget    w;
XButtonEvent   *event;
{
    int  next_select_char;

    next_select_char = pixel_coords_to_char(event->x, event->y);

    if (next_select_char == last_drawn)
	return;

    if (done_first) {
	draw_range(XtWindow(w), last_drawn, first_select_char, normal_gc);
	draw_range(XtWindow(w), first_select_char, next_select_char, 
		   inverse_gc);
    } else {
	draw_range(XtWindow(w), first_select_char, next_select_char, 
		   inverse_gc);
	done_first = True;
    }

    last_drawn = next_select_char;
}


/*
    Btn1 has been released. Remember the character the pointer was
    over when the button was released. 

    Set first_select_char and last_select_char to point to the range
    of selected text. The text is inclusive of first_select_char and
    exclusive of last_select_char.

    Assert ownership of the primary selection.
*/

void end_select (w, event)
Widget    w;
XButtonEvent   *event;
{
    int tmp;

    last_select_char = pixel_coords_to_char(event->x, event->y);

    /*
       We didn't get any text selected.
    */
    if (last_select_char == first_select_char) {
	/* Disown PRIMARY selection if we own it */
	if (own_primary)
	    XtDisownSelection(w, xa_PRIMARY, event->time);
	first_select_char = last_select_char = -1;
	return;
    }

    /* Get them in order */
    if (last_select_char < first_select_char) {
	tmp = last_select_char;
	last_select_char = first_select_char;
	first_select_char = tmp;
    }

    first_select_char = MAX(0,first_select_char);
    last_select_char = MIN(strlen(the_text), last_select_char);

    /*
       Okay, now we've got something selected. Let's tell the world.
    */
    own_primary = XtOwnSelection(w, xa_PRIMARY, event->time, 
				 selection_requestCB,
				 lost_ownershipCB,
				 NULL);
    /*
       Have to make sure we actually obtained ownership.
       If we didn't, take steps to clear selection data.
    */
    if (!own_primary)
	lost_ownershipCB(w, xa_PRIMARY);

}
	


/***************************************************************

 Stuff for Managing Selection Ownership and Requests. (Cutting)

***************************************************************/


/*
    Someone has asked us for the value of a selection.
    Respond appropriately, based on the target they specified.
*/

Boolean selection_requestCB (w, selection, target, type, value, length, format)
Widget	w;	     	/* In */
Atom	*selection;	/* In */
Atom	*target;	/* In */
Atom	*type;		/* Out */
XtPointer *value;	/* Out */
unsigned long *length;	/* Out */
int	*format;	/* Out */
{
     int   i;
     XSelectionRequestEvent   *req;
     Atom	*tmp_target_list;
     int	*tmp_length;

#ifdef DEBUG
     printf("In requestCB, target is %s\n", XGetAtomName(dpy, *target));
#endif

     req = XtGetSelectionRequest(w, *selection, NULL);

    /*
       First, handle requests for TARGETS
    */
    if (*target == xa_TARGETS) {
	Atom   *xmu_targets;
	unsigned long  xmu_length;

	/* 
	   Get Xmu to list targets it handles 
	*/
	XmuConvertStandardSelection(w, req->time, selection, target, type,
				    &xmu_targets, &xmu_length, format);
	/*
	    Put our targets in the list.
	*/
	*length = xmu_length + 3; /* 3 is number of targets we support */
	tmp_target_list = (Atom *)XtMalloc(sizeof(Atom) * (*length));
	tmp_target_list[0] = xa_STRING;
	tmp_target_list[1] = xa_TEXT;
	tmp_target_list[2] = xa_LENGTH;
	/*
	    Copy the Xmu targets into our list
	*/
	for (i = 3; i < *length; i++)
	    tmp_target_list[i] = xmu_targets[i-3];
	XtFree((char *)xmu_targets);
	*value = (XtPointer)tmp_target_list;
	*type = xa_ATOM;
	*format = 32;
	return True;
    }

    /*
	Handle a request for STRING or TEXT
	Malloc space and copy the string in.
	Just pass a pointer to the appropriate part of the string. 
    */
    if ((*target == xa_TEXT) || (*target == xa_STRING)) {
	*length = last_select_char - first_select_char;
	*value = XtMalloc(*length);
	for (i = 0; i < *length; i++)
	    (*value)[i] = the_text[first_select_char + i];
	*type = xa_STRING;
	*format = 8;
	return True;
    }

    /*
	Handle a request for LENGTH
    */
    if ((*target == xa_LENGTH)) {
	tmp_length = (int *)XtMalloc(sizeof(int));
	*tmp_length = last_select_char - first_select_char;
	*value = (XtPointer)tmp_length;
	*length = 1;
	*type = xa_INTEGER;
	*format = sizeof(int) * 8;
	return True;
    }

    /*
	We couldn't do it, maybe Xmu can
    */
    return (XmuConvertStandardSelection(w, req->time, selection, target,
				        type, value, length, format));

}


void lost_ownershipCB (w, selection)
Widget   w;
Atom     *selection;
{

    /*
	If we lose ownership of primary, clean up our highlight and
	reset our state.
    */
    if (*selection == xa_PRIMARY) {
    	draw_range(XtWindow(w), first_select_char, last_select_char, normal_gc);
    	done_first = False;
    	first_select_char = last_select_char = -1;
    	last_drawn = -1;
	own_primary = False;
    }
}



/*******************************************************************

 Stuff for Obtaining Selection Values Upon User Request. (Pasting)

*******************************************************************/


/*
    User asked to get PRIMARY selection.
    First find out what formats are available.
*/

void paste_primary (w, event)
Widget   w;
XButtonEvent	*event;
{
    XtGetSelectionValue (w, xa_PRIMARY, xa_TARGETS, targetCB, 
			 (XtPointer)event->time, event->time);
}


/*
    Pick the best format from the list available and ask
    for a conversion to that target.
*/

void targetCB (w, client_data, selection, type, value, length, format)
Widget   	w;
XtPointer 	client_data;
Atom     	*selection;
Atom  	 	*type;
XtPointer	value;
unsigned long	*length;
int		*format;
{
    Atom    best_target;
    Time    paste_time = (Time)client_data;
     
    if (*type == XT_CONVERT_FAIL) {
	  error_dialog("Paste operation timed out, try again.");
	  return;
    }

    if ((*type == None) || (*length == 0)) {
	/* Could be that owner doesn't support TARGETS, just try STRING */
	best_target = xa_STRING;
    } else {
	best_target = pick_best_target((Atom *)value, *length);
    }

    if (best_target == NULL)
	error_dialog("No appropriate data to paste");
    else
	XtGetSelectionValue (w, *selection, best_target, valueCB, 
			     NULL, paste_time);
}


/*
    We got the data, let's tack it on to our string.
*/

void valueCB (w, client_data, selection, type, value, length, format)
Widget   	w;
XtPointer 	client_data;
Atom     	*selection;
Atom  	 	*type;
XtPointer	value;
unsigned long	*length;
int		*format;
{
    int   current_length;
    int	  chars_to_copy, i;
    char  *tmp_ptr;

    /*
	Check various possibilities for error.
    */

    if (*type == XT_CONVERT_FAIL) {
	error_dialog("Paste operation timed out, try again.");
	return;
    }
    if (*type == None) {
	error_dialog("No selection, or Paste operation failed.");
	return;
    }
    if ((*type != xa_STRING) && (*type == xa_TEXT)) {
	XtFree(value);
	error_dialog("Attempt to paste unsupported data");
	return;
    }

    current_length = strlen(the_text);
    chars_to_copy = MIN(*length, MAX_TEXT_LEN - current_length);
    tmp_ptr = the_text + current_length;
    for (i = 0; i < chars_to_copy; i++) {
	*tmp_ptr++ = *(value + i);
    }
    *tmp_ptr = '\0';

    /* Draw the new part of our string */
    draw_range(XtWindow(w), current_length, current_length + chars_to_copy + 1,
	       normal_gc);

    /* Let the user know if we dropped any characters */
    if ((*length + current_length) > MAX_TEXT_LEN) {
	error_dialog("Complete paste would've exceeded maximum buffer size.\
 Partial paste performed.");
    }

    XtFree(value);
}


/**************************************************************

		       Utilities

**************************************************************/

static Atom  *preferred_targets[] = {
				   &xa_STRING,
				   &xa_TEXT
				   };

/*
    Go through our list of preferred targets and see if
    they are supported by the owning client. Return the
    first match.

    If none match, return NULL.
*/

Atom pick_best_target (target_list, n)
Atom  *target_list;
unsigned long   n;
{
    int i, j;

    for (i = 0; i < XtNumber(preferred_targets); i++)
	for (j = 0; j < n; j++)
	    if (target_list[j] == *preferred_targets[i])
		return (target_list[j]);

    return NULL;
}


/*
    Put up a dialog with the given message.
*/

void error_dialog (str)
char  *str;
{
    Widget    dialog;
    Arg	      arg[5];
    XmString  tmpxmstr;
    int	      n;

    n = 0;
    tmpxmstr = XmStringCreateSimple(str);
    XtSetArg(arg[n], XmNmessageString, tmpxmstr); n++;
    XtSetArg(arg[n], XmNdeleteResponse, XmDESTROY); n++;
    dialog = XmCreateMessageDialog(toplevel, "error_dialog", arg, n);
    XmStringFree(tmpxmstr);

    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
    XtManageChild(dialog);

}

